@@ -77,7 +77,7 @@ class FilenameEditor(QWidget):
 
         orig = default = self._line_edit.text()
         if default != '' and self._project is not None:
-            default = self._project.absolute_path(default)
+            default = self._project.path_from_user(default)
 
         if self._directory:
             name = QFileDialog.getExistingDirectory(self._line_edit,
@@ -87,10 +87,7 @@ class FilenameEditor(QWidget):
                     self._caption, default)
 
         if name != '':
-            name = QDir.toNativeSeparators(name)
-
-            if self._project is not None:
-                name = self._project.relative_path(name)
+            name = self._project.path_to_user(name)
 
             if name != orig:
                 self._line_edit.setText(name)
