@@ -25,5 +25,7 @@
 
 
 # Publish the sub-package's API.
+from .external_libs_metadata import external_libraries_metadata
 from .pyqt4 import pyqt4_metadata
 from .pyqt5 import pyqt5_metadata
+from .python_metadata import ExtensionModule, get_python_metadata
