@@ -27,21 +27,17 @@
 import os
 from xml.etree.ElementTree import Element, ElementTree, SubElement
 
-from PyQt5.QtCore import QObject, pyqtSignal
+from PyQt5.QtCore import QDir, QFileInfo, QObject, pyqtSignal
 
+from ..metadata import get_python_metadata
 from ..user_exception import UserException
 
 
 class Project(QObject):
     """ The encapsulation of a project. """
 
-    # The current project version.
-    #   0: The original version.
-    #   1: Added the Others element with the builddir and qmake attributes.
-    #   2: The SitePackages element can now contain multiple Package
-    #      sub-elements.
-    #      Added the syspath attribute to the Application element.
-    version = 2
+    # The current project version.  Versions 0 to 3 are no longer supported.
+    version = 4
 
     # Emitted when the modification state of the project changes.
     modified_changed = pyqtSignal(bool)
@@ -67,43 +63,17 @@ class Project(QObject):
     def name(self):
         """ The name property getter. """
 
-        return self._name
+        # Use absoluteFilePath() because the file might not exist.
+        return self._name.absoluteFilePath() if self._name is not None else ''
 
     @name.setter
     def name(self, value):
         """ The name property setter. """
 
-        if self._name != value:
-            self._name = value
+        if self._name is None or self._name.absoluteFilePath() != value:
+            self._name = QFileInfo(value)
             self.name_changed.emit(value)
 
-    @property
-    def python_target_version(self):
-        """ The target Python version getter. """
-
-        if self.python_target_library != '':
-            lib_name = os.path.basename(self.python_target_library)
-
-            # Strip everything up to the leading digit.
-            while lib_name != '' and not lib_name[0].isdigit():
-                lib_name = lib_name[1:]
-
-            # Strip everything after the trailing digit.
-            while lib_name != '' and not lib_name[-1].isdigit():
-                lib_name = lib_name[:-1]
-
-            version_str = lib_name.replace('.', '')
-
-            # We should now have a 2-digit string.
-            try:
-                version = int(version_str)
-
-                return version // 10, version % 10
-            except ValueError:
-                pass
-
-        return None, None
-
     def __init__(self, name=''):
         """ Initialise the project. """
 
@@ -111,73 +81,168 @@ class Project(QObject):
 
         # Initialise the project meta-data.
         self._modified = False
-        self._name = os.path.abspath(name) if name != '' else ''
+        self._name = QFileInfo(name) if name != '' else None
 
         # Initialise the project data.
+        self.application_name = ''
         self.application_is_pyqt5 = True
+        self.application_is_console = False
+        self.application_is_bundle = True
         self.application_package = QrcPackage()
         self.application_script = ''
+        self.application_entry_point = ''
         self.build_dir = 'build'
-        self.extension_modules = []
+        self.external_libraries = []
+        self.other_extension_modules = []
+        self.other_packages = []
         self.pyqt_modules = []
         self.python_host_interpreter = ''
+        self.python_source_dir = ''
+        self.python_ssl = False
         self.python_target_include_dir = ''
         self.python_target_library = ''
         self.python_target_stdlib_dir = ''
+        self.python_target_version = (3, 4)
         self.qmake = ''
-        self.packages = [QrcPackage()]
-        self.stdlib_package = QrcPackage()
+        self.standard_library = []
         self.sys_path = ''
 
-    def relative_path(self, filename):
-        """ Convert a filename to one that is relative to the project
-        name if possible.
+    def path_to_user(self, path):
+        """ Convert a file name to one that is relative to the project name if
+        possible and uses native separators.
         """
 
-        filename = filename.strip()
+        if self._name is not None:
+            rel = self._name.dir().relativeFilePath(path)
+            if not rel.startswith('..'):
+                path = rel
 
-        if os.path.isabs(filename):
-            if self._name != '':
-                project_dir = os.path.dirname(self._name)
-                if filename.startswith(project_dir):
-                    filename = os.path.relpath(filename, project_dir)
+        return QDir.toNativeSeparators(path)
 
-        return filename
-
-    def absolute_path(self, filename):
-        """ Convert a filename that might be relative to the project name to an
-        absolute filename.
+    def path_from_user(self, user_path):
+        """ Convert the name of a file or directory specified by the user to
+        the standard Qt format (ie. an absolute path using UNIX separators).  A
+        user path may be relative to the name of the project and may contain
+        environment variables.
         """
 
-        filename = os.path.expandvars(filename.strip())
+        fi = self._fileinfo_from_user(user_path)
 
-        if not os.path.isabs(filename):
-            filename = os.path.join(os.path.dirname(self._name), filename)
+        # Use the canonical name if possible (ie. when the file exists) and
+        # fall back to the absolute name.
+        path = fi.canonicalFilePath()
+        if path == '':
+            path = fi.absoluteFilePath()
 
-        return filename
+        return path
 
-    def application_basename(self):
-        """ Return the basename of the application script (i.e. with no path or
-        extension.
+    def get_executable_basename(self):
+        """ Return the basename of the application executable (i.e. with no
+        path or extension.
         """
 
-        app_name, _ = os.path.splitext(
-                os.path.basename(self.absolute_path(self.application_script)))
+        if self.application_name != '':
+            return self.application_name
 
-        return app_name
+        name = self.application_script
+        if name == '':
+            name = self.application_package.name
+            if name == '':
+                return ''
+
+        return self._fileinfo_from_user(name).completeBaseName()
+
+    def _fileinfo_from_user(self, user_path):
+        """ Convert the name of a file or directory specified by the user to a
+        QFileInfo instance.  A user path may be relative to the name of the
+        project and may contain environment variables.
+        """
+
+        fi = QFileInfo(os.path.expandvars(user_path.strip()))
+
+        if fi.isRelative():
+            fi = QFileInfo(self._name.canonicalPath() + '/' + fi.filePath())
+
+        return fi
+
+    def get_stdlib_requirements(self, include_hidden=False):
+        """ Return a 2-tuple of the required Python standard library modules
+        and the required external libraries.  The modules are a dict with the
+        module name as the key and a bool as the value.  The bool is True if
+        the module is explicitly required and False if it is implicitly
+        required.  The libraries are a set of well known library names.
+        """
+
+        # Work out the dependencies.
+        metadata = get_python_metadata(self.python_target_version)
+        all_modules = {name: _DepState(module)
+                for name, module in metadata.items()}
+
+        visit = 0
+        for name in all_modules.keys():
+            self._set_dependency_state(all_modules, name, visit)
+            visit += 1
+
+        # Extract the required modules and libraries.
+        required_modules = {}
+        required_libraries = set()
+
+        for name, dep_state in all_modules.items():
+            if dep_state.explicit:
+                explicit = True
+            elif dep_state.implicit:
+                explicit = False
+            else:
+                continue
+
+            # Handle any hidden dependencies if required.
+            if include_hidden:
+                for hidden_dep in dep_state.module.hidden_deps:
+                    if hidden_dep not in required_modules:
+                        required_modules[hidden_dep] = False
+
+            required_modules[name] = explicit
+
+            if dep_state.module.xlib is not None:
+                required_libraries.add(dep_state.module.xlib)
+
+        return required_modules, required_libraries
+
+    def _set_dependency_state(self, all_modules, name, visit, is_dep=False):
+        """ Set a module's dependency state. """
+
+        dep_state = all_modules[name]
+
+        if dep_state.visit == visit:
+            return
+
+        dep_state.visit = visit
+
+        if dep_state.module.ssl is not None:
+            if dep_state.module.ssl != self.python_ssl:
+                return
+
+        dep_state.explicit = (name in self.standard_library)
+
+        if dep_state.module.core or is_dep:
+            dep_state.implicit = True
+
+        for dep in dep_state.module.deps:
+            self._set_dependency_state(all_modules, dep, visit,
+                    (dep_state.explicit or dep_state.implicit))
 
     @classmethod
-    def load(cls, filename):
+    def load(cls, file_name):
         """ Return a new project loaded from the given file.  Raise a
         UserException if there was an error.
         """
 
-        abs_filename = os.path.abspath(filename)
+        fi = QFileInfo(file_name)
 
         tree = ElementTree()
 
         try:
-            root = tree.parse(abs_filename)
+            root = tree.parse(QDir.toNativeSeparators(fi.canonicalFilePath()))
         except Exception as e:
             raise UserException(
                 "There was an error reading the project file.", str(e))
@@ -197,70 +262,100 @@ class Project(QObject):
 
         cls._assert(version is not None, "Invalid 'version'.")
 
+        if version <= 3:
+            raise UserException("The project's format is no longer supported.")
+
         if version > cls.version:
             raise UserException(
                     "The project's format is version {0} but only version {1} is supported.".format(version, cls.version))
 
         # Create the project and populate it.
         project = cls()
-        project.name = abs_filename
-
-        # The application specific configuration.
-        application = root.find('Application')
-        cls._assert(application is not None, "Missing 'Application' tag.")
-
-        project.application_is_pyqt5 = cls._get_bool(application, 'ispyqt5',
-                'Application')
-        project.application_script = application.get('script', '')
-        project.sys_path = application.get('syspath', '')
-
-        app_package = application.find('Package')
-        cls._assert(app_package is not None,
-                "Missing 'Application.Package' tag.")
-        project.application_package = cls._load_package(app_package)
-
-        for pyqt_m in application.iterfind('PyQtModule'):
-            name = pyqt_m.get('name', '')
-            cls._assert(name != '',
-                    "Missing or empty 'PyQtModule.name' attribute.")
-            project.pyqt_modules.append(name)
-
-        site_packages = application.find('SitePackages')
-        cls._assert(site_packages is not None,
-                "Missing 'Application.SitePackages' tag.")
-        project.packages = [cls._load_package(package)
-                for package in site_packages.iterfind('Package')]
-        cls._assert(len(project.packages) != 0,
-                "Missing 'Application.SitePackages.Package' tag.")
-
-        stdlib = application.find('Stdlib')
-        cls._assert(stdlib is not None, "Missing 'Application.Stdlib' tag.")
-        stdlib_package = stdlib.find('Package')
-        cls._assert(stdlib_package is not None,
-                "Missing 'Application.Stdlib.Package' tag.")
-        project.stdlib_package = cls._load_package(stdlib_package)
-
-        for extension_module_element in application.iterfind('ExtensionModule'):
-            name = extension_module_element.get('name')
-            cls._assert(name is not None,
-                    "Missing 'Application.ExtensionModule.name' attribute.")
-
-            path = extension_module_element.get('path')
-            cls._assert(path is not None,
-                    "Missing 'Application.ExtensionModule.path' attribute.")
-
-            project.extension_modules.append(ExtensionModule(name, path))
+        project._name = fi
 
         # The Python specific configuration.
         python = root.find('Python')
         cls._assert(python is not None, "Missing 'Python' tag.")
 
         project.python_host_interpreter = python.get('hostinterpreter', '')
+        project.python_source_dir = python.get('sourcedir', '')
+        project.python_ssl = cls._get_bool(python, 'ssl', 'Python')
         project.python_target_include_dir = python.get('targetincludedir', '')
         project.python_target_library = python.get('targetlibrary', '')
         project.python_target_stdlib_dir = python.get('targetstdlibdir', '')
 
-        # The other configuration (added in version 1).
+        major = cls._get_int(python, 'major', 'Python')
+        minor = cls._get_int(python, 'minor', 'Python')
+        project.python_target_version = (major, minor)
+
+        # The application specific configuration.
+        application = root.find('Application')
+        cls._assert(application is not None, "Missing 'Application' tag.")
+
+        project.application_entry_point = application.get('entrypoint', '')
+        project.application_is_pyqt5 = cls._get_bool(application, 'ispyqt5',
+                'Application')
+        project.application_is_console = cls._get_bool(application,
+                'isconsole', 'Application')
+        project.application_is_bundle = cls._get_bool(application, 'isbundle',
+                'Application')
+        project.application_name = application.get('name', '')
+        project.application_script = application.get('script', '')
+        project.sys_path = application.get('syspath', '')
+
+        # Any application package.
+        app_package = application.find('Package')
+
+        if app_package is not None:
+            project.application_package = cls._load_package(app_package)
+        else:
+            project.application_package = QrcPackage()
+
+        # Any PyQt modules.
+        for pyqt_m in root.iterfind('PyQtModule'):
+            name = pyqt_m.get('name', '')
+            cls._assert(name != '',
+                    "Missing or empty 'PyQtModule.name' attribute.")
+            project.pyqt_modules.append(name)
+
+        # Any standard library modules.
+        for stdlib_module_element in root.iterfind('StdlibModule'):
+            name = stdlib_module_element.get('name')
+            cls._assert(name is not None,
+                    "Missing 'StdlibModule.name' attribute.")
+
+            project.standard_library.append(name)
+
+        # Any external C libraries.
+        for external_lib_element in root.iterfind('ExternalLib'):
+            name = external_lib_element.get('name')
+            cls._assert(name is not None,
+                    "Missing 'ExternalLib.name' attribute.")
+
+            defines = external_lib_element.get('defines', '')
+            includepath = external_lib_element.get('includepath', '')
+            libs = external_lib_element.get('libs', '')
+
+            project.external_libraries.append(
+                    ExternalLibrary(name, defines, includepath, libs))
+
+        # Any other Python packages.
+        project.other_packages = [cls._load_package(package)
+                for package in root.iterfind('Package')]
+
+        # Any other extension module.
+        for extension_module_element in root.iterfind('ExtensionModule'):
+            name = extension_module_element.get('name')
+            cls._assert(name is not None,
+                    "Missing 'ExtensionModule.name' attribute.")
+
+            libs = extension_module_element.get('libs')
+            cls._assert(path is not None,
+                    "Missing 'ExtensionModule.libs' attribute.")
+
+            project.other_extension_modules.append(ExtensionModule(name, libs))
+
+        # The other configuration.
         others = root.find('Others')
         if others is not None:
             project.build_dir = others.get('builddir', '')
@@ -273,18 +368,16 @@ class Project(QObject):
 
         self._save_project(self.name)
 
-    def save_as(self, filename):
+    def save_as(self, file_name):
         """ Save the project to the given file and make the file the
         destination of subsequent saves.  Raise a UserException if there was an
         error.
         """
 
-        abs_filename = os.path.abspath(filename)
-
-        self._save_project(abs_filename)
+        self._save_project(file_name)
 
         # Only do this after the project has been successfully saved.
-        self.name = abs_filename
+        self.name = file_name
 
     @classmethod
     def _load_package(cls, package_element):
@@ -349,7 +442,23 @@ class Project(QObject):
 
         return bool(value)
 
-    def _save_project(self, abs_filename):
+    @classmethod
+    def _get_int(cls, element, name, context):
+        """ Get an integer attribute from an element. """
+
+        value = element.get(name)
+        try:
+            value = int(value)
+        except:
+            value = None
+
+        cls._assert(value is not None,
+                "Missing or invalid integer value of '{0}.{1}'.".format(
+                        context, name))
+
+        return value
+
+    def _save_project(self, file_name):
         """ Save the project to the given file.  Raise a UserException if there
         was an error.
         """
@@ -357,34 +466,50 @@ class Project(QObject):
         root = Element('Project', attrib={
             'version': str(self.version)})
 
+        SubElement(root, 'Python', attrib={
+            'hostinterpreter': self.python_host_interpreter,
+            'sourcedir': self.python_source_dir,
+            'ssl': str(int(self.python_ssl)),
+            'targetincludedir': self.python_target_include_dir,
+            'targetlibrary': self.python_target_library,
+            'targetstdlibdir': self.python_target_stdlib_dir,
+            'major': str(self.python_target_version[0]),
+            'minor': str(self.python_target_version[1])})
+
         application = SubElement(root, 'Application', attrib={
+            'entrypoint': self.application_entry_point,
             'ispyqt5': str(int(self.application_is_pyqt5)),
+            'isconsole': str(int(self.application_is_console)),
+            'isbundle': str(int(self.application_is_bundle)),
+            'name': self.application_name,
             'script': self.application_script,
             'syspath': self.sys_path})
 
-        self._save_package(application, self.application_package)
+        if self.application_package.name != '':
+            self._save_package(application, self.application_package)
 
         for pyqt_module in self.pyqt_modules:
-            SubElement(application, 'PyQtModule', attrib={
+            SubElement(root, 'PyQtModule', attrib={
                 'name': pyqt_module})
 
-        site_packages = SubElement(application, 'SitePackages')
-        for package in self.packages:
-            self._save_package(site_packages, package)
+        for stdlib_module in self.standard_library:
+            SubElement(root, 'StdlibModule', attrib={
+                'name': stdlib_module})
 
-        stdlib = SubElement(application, 'Stdlib')
-        self._save_package(stdlib, self.stdlib_package)
+        for external_lib in self.external_libraries:
+            SubElement(root, 'ExternalLib', attrib={
+                'name': external_lib.name,
+                'defines': external_lib.defines,
+                'includepath': external_lib.includepath,
+                'libs': external_lib.libs})
 
-        for extension_module in self.extension_modules:
-            SubElement(application, 'ExtensionModule', attrib={
+        for package in self.other_packages:
+            self._save_package(root, package)
+
+        for extension_module in self.other_extension_modules:
+            SubElement(root, 'ExtensionModule', attrib={
                 'name': extension_module.name,
-                'path': extension_module.path})
-
-        SubElement(root, 'Python', attrib={
-            'hostinterpreter': self.python_host_interpreter,
-            'targetincludedir': self.python_target_include_dir,
-            'targetlibrary': self.python_target_library,
-            'targetstdlibdir': self.python_target_stdlib_dir})
+                'libs': extension_module.libs})
 
         SubElement(root, 'Others', attrib={
             'builddir': self.build_dir,
@@ -393,7 +518,8 @@ class Project(QObject):
         tree = ElementTree(root)
 
         try:
-            tree.write(abs_filename, encoding='unicode', xml_declaration=True)
+            tree.write(QDir.toNativeSeparators(file_name), encoding='unicode',
+                    xml_declaration=True)
         except Exception as e:
             raise UserException(
                     "There was an error writing the project file.", str(e))
@@ -493,11 +619,37 @@ class QrcDirectory(QrcFile):
         return copy
 
 
+class ExternalLibrary():
+    """ The encapsulation of an external library. """
+
+    def __init__(self, name, defines, includepath, libs):
+        """ Initialise the external library. """
+
+        self.name = name
+        self.defines = defines
+        self.includepath = includepath
+        self.libs = libs
+
+
 class ExtensionModule():
     """ The encapsulation of an extension module. """
 
-    def __init__(self, name, path):
-        """ Initialise the extension. """
+    def __init__(self, name, libs):
+        """ Initialise the extension module. """
 
         self.name = name
-        self.path = path
+        self.libs = libs
+
+
+class _DepState:
+    """ Encapsulate the state information required when working out module
+    dependencies.
+    """
+
+    def __init__(self, module):
+        """ Initialise the object. """
+
+        self.module = module
+        self.explicit = False
+        self.implicit = False
+        self.visit = -1
